from .BasePlayer import BasePlayer
from .consts import MPyg321Events, PlayerStatus
from .EventContext import MPyg321EventContext


class MPyg123Player(BasePlayer):
    """Player for mpg123"""

    def __init__(
        self,
        player=None,
        audiodevice=None,
        performance_mode=True,
        custom_args="",
        rva_mix=False,
    ):
        self.suitable_versions = ["mpg123"]
        self.default_player = "mpg123"
        custom_args += " --rva-mix " if rva_mix else ""
        super().__init__(player, audiodevice, performance_mode, custom_args)
        if performance_mode:
            self.silence_mpyg_output()
        self._is_muted = False            

    def process_output_ext(self, action):
        """Processes specific output for mpg123 player"""
        if action == "user_mute":
            self._is_muted = True
            self.on_user_mute()
            self._trigger_event(MPyg321Events.USER_MUTE, MPyg321EventContext(self))
        elif action == "user_unmute":
            self._is_muted = False
            self._trigger_event(MPyg321Events.USER_UNMUTE, MPyg321EventContext(self))
            self.on_user_unmute()

    def load_list(self, entry, filepath):
        """Load an entry in a list
        Parameters:
        entry (int): index of the song in the list - first is 0
        filepath: URL/Path to the list
        """
        self.player.sendline("LOADLIST {} {}".format(entry, filepath))
        self.status = PlayerStatus.PLAYING

    def silence_mpyg_output(self):
        """Improves performance by silencing the mpg123 process frame output"""
        self.player.sendline("SILENCE")

    def mute(self):
        """Mutes the player"""
        self.player.sendline("MUTE")

    def unmute(self):
        """Unmutes the player"""
        self.player.sendline("UNMUTE")

    def toggle_mute(self):
        """Mute or UnMute if playing"""
        if self._is_muted:
            self.unmute()
        else:
            self.mute()

    def volume(self, percent):
        """Adjust player's volume"""
        self.player.sendline("VOLUME {}".format(percent))

    # # # Public Callbacks # # #
    def on_user_mute(self):
        """Callback when user mutes player"""
        pass

    def on_user_unmute(self):
        """Callback when user unmutes player"""
        pass



"""
This file is here for backwards compatibility with previous versions
So that users can use: from mpyg321.mpyg321 import MPyg321Player
Newer version would be: from mpyg321.MPyg321Player import Mpyg321Player
"""
from .MPyg123Player import MPyg123Player
from .MPyg321Player import MPyg321Player



class MPyg321EventContext:
    """Base class for all events"""

    def __init__(self, player) -> None:
        self.player = player


class MPyg321ErrorContext(MPyg321EventContext):
    """Context for error events"""

    def __init__(self, player, error_type, error_message) -> None:
        super().__init__(player)
        self.error_type = error_type
        self.error_message = error_message



from enum import Enum

from pexpect import TIMEOUT as pexpectTIMEOUT

mpg_outs = [
    {
        "mpg_code": "@P 0",
        "action": "music_stop",
        "description": """For mpg123, it corresponds to any stop
                        For mpg312 it corresponds to user stop only""",
    },
    {
        "mpg_code": "@P 1",
        "action": "user_pause",
        "description": "Music has been paused by the user.",
    },
    {
        "mpg_code": "@P 2",
        "action": "user_start_or_resume",
        "description": "Music has been started resumed by the user.",
    },
    {
        "mpg_code": "@E *",
        "action": "error",
        "description": "Player has encountered an error.",
    },
    {
        "mpg_code": "@silence",
        "action": None,
        "description": "Player has been silenced by the user.",
    },
    {
        "mpg_code": r"@V [0-9\.\s%]*",
        "action": None,
        "description": "Volume change event.",
    },
    {
        "mpg_code": r"@S [a-zA-Z0-9\.\s-]*",
        "action": None,
        "description": "Stereo info event.",
    },
    {"mpg_code": "@I *", "action": None, "description": "Information event."},
    {"mpg_code": pexpectTIMEOUT, "action": None, "description": "Timeout event."},
]

mpg_outs_ext = {
    "mpg123": [
        {
            "mpg_code": "@mute",
            "action": "user_mute",
            "description": "Player has been muted by the user.",
        },
        {
            "mpg_code": "@unmute",
            "action": "user_unmute",
            "description": "Player has been unmuted by the user.",
        },
    ],
    "mpg321": [
        {
            "mpg_code": "@P 3",
            "action": "end_of_song",
            "description": "Player has reached the end of the song.",
        }
    ],
}

mpg_errors = [
    {"message": "empty list name", "action": "generic_error"},
    {"message": "No track loaded!", "action": "generic_error"},
    {"message": "Error opening stream", "action": "file_error"},
    {"message": "failed to parse given eq file:", "action": "file_error"},
    {"message": "Corrupted file:", "action": "file_error"},
    {"message": "Unknown command:", "action": "command_error"},
    {"message": "Unfinished command:", "action": "command_error"},
    {"message": "Unknown command or no arguments:", "action": "argument_error"},
    {"message": "invalid arguments for", "action": "argument_error"},
    {"message": "Missing argument to", "action": "argument_error"},
    {"message": "failed to set eq:", "action": "eq_error"},
    {"message": "Error while seeking", "action": "seek_error"},
]


class PlayerStatus:
    INSTANCIATED = 0
    PLAYING = 1
    PAUSED = 2
    RESUMING = 3
    STOPPING = 4
    STOPPED = 5
    QUITTED = 6


class MPyg321Events(Enum):
    USER_STOP = "user_stop"
    USER_PAUSE = "user_pause"
    USER_RESUME = "user_resume"
    ANY_STOP = "any_stop"
    MUSIC_END = "music_end"
    ERROR = "error"
    USER_MUTE = "mute"
    USER_UNMUTE = "unmute"



# # # Errors # # #
class MPygError(RuntimeError):
    """Base class for any errors encountered by the player during runtime"""

    pass


class MPygFileError(MPygError):
    """Errors encountered by the player related to files"""

    pass


class MPygCommandError(MPygError):
    """Errors encountered by the player related to player commands"""

    pass


class MPygArgumentError(MPygError):
    """Errors encountered by the player related to arguments for commands"""

    pass


class MPygEQError(MPygError):
    """Errors encountered by the player related to the equalizer"""

    pass


class MPygSeekError(MPygError):
    """Errors encountered by the player related to the seek"""

    pass


class MPygPlayerNotFoundError(MPygError):
    """Errors encountered when no suitable player is found"""

    pass


class MPygUnknownEventNameError(MPygError):
    """Errors encountered when creating an event listener for a non existing event"""

    pass


class MPygEventListenerError(MPygError):
    """Errors encountered when an event listener throws an exception"""

    pass



"""
Mpyg BasePlayer base class
This class contains all the functions that are common
both to mpg321 and mpg123.
All the players implement this base class and add their
specific feature.
"""

import subprocess
from threading import Thread

import pexpect

from .consts import *
from .EventContext import *
from .MpygError import *


class BasePlayer:
    """Base class for players"""

    player = None
    status = None
    output_processor = None
    song_path = ""
    loop = False
    performance_mode = True
    suitable_versions = []  # mpg123 and/or mpg321 - set inside subclass
    default_player = None  # mpg123 or mpg321 - set inside subclass
    player_version = None  # defined inside check_player
    mpg_outs = []
    _events = {}

    def __init__(
        self, player=None, audiodevice=None, performance_mode=True, custom_args=""
    ):
        """Builds the player and creates the callbacks"""
        self._events = {e: [] for e in MPyg321Events}
        self.set_player(player, audiodevice, custom_args)
        self.output_processor = Thread(target=self.process_output)
        self.output_processor.daemon = True
        self.performance_mode = performance_mode
        self.output_processor.start()

    def check_player(self, player):
        """Gets the player"""
        try:
            cmd = str(player)
            output = subprocess.check_output([cmd, "--version"])
            for version in self.suitable_versions:
                if version in str(output):
                    self.player_version = version
            if self.player_version is None:
                raise MPygPlayerNotFoundError(
                    """No suitable player found: you might be using the wrong \
player (Mpyg321Player or Mpyg123Player)"""
                )
        except subprocess.SubprocessError:
            raise MPygPlayerNotFoundError(
                """No suitable player found: you might need to install
                mpg123"""
            )

    def set_player(self, player, audiodevice, custom_args):
        """Sets the player"""
        if player is None:
            player = self.default_player
        self.check_player(player)
        args = " " + custom_args if custom_args != "" else ""
        args += " --audiodevice " + audiodevice if audiodevice else ""
        args += " -R mpyg"
        self.player = pexpect.spawn(str(player) + " " + args)
        self.player.delaybeforesend = None
        self.status = PlayerStatus.INSTANCIATED
        # Setting extended mpg_outs for version specific behaviors
        self.mpg_outs = mpg_outs.copy()
        self.mpg_outs.extend(mpg_outs_ext[self.player_version])

    def on(self, event_name):
        """Decorator to register event callbacks."""

        def decorator(func):
            if event_name not in self._events:
                raise MPygUnknownEventNameError(
                    f"Subscribed callback to a non existing event {event_name}."
                )
            self._events[event_name].append(func)
            return func

        return decorator

    def subscribe_event(self, event_name, callback):
        if event_name not in self._events:
            raise MPygUnknownEventNameError(
                f"Subscribed callback to a non existing event {event_name}."
            )
        self._events[event_name].append(callback)

    def _trigger_event(self, event_name, context=None):
        """Trigger all callbacks associated with an event."""
        if context is None:
            context = MPyg321EventContext(self)
        if event_name in self._events:
            for callback in self._events[event_name]:
                try:
                    callback(context)
                except Exception:
                    raise MPygEventListenerError(
                        "Error while executiong event callback"
                    )

    def process_output(self):
        """Parses the output"""
        mpg_codes = [v["mpg_code"] for v in self.mpg_outs]
        while True:
            index = self.player.expect(mpg_codes)
            action = self.mpg_outs[index]["action"]
            if action == "music_stop":
                self.on_music_stop_int()
            elif action == "user_pause":
                self.on_user_pause_int()
            elif action == "user_start_or_resume":
                self.on_user_start_or_resume_int()
            elif action == "end_of_song":
                self.on_end_of_song_int()
            elif action == "error":
                self.on_error()
            else:
                self.process_output_ext(action)

    def process_output_ext(self, action):
        """Processes the output for version specific behavior"""
        pass

    def play_song(self, path, loop=False):
        """Plays the song"""
        self.loop = loop
        self.set_song(path)
        self.play()

    def play(self):
        """Starts playing the song"""
        self.player.sendline("LOAD " + self.song_path)
        self.status = PlayerStatus.PLAYING

    def pause(self):
        """Pauses the player"""
        if self.status == PlayerStatus.PLAYING:
            self.player.sendline("PAUSE")
            self.status = PlayerStatus.PAUSED

    def toggle_pause(self):
        """Pause if playing, else resume if paused"""
        if self.status == PlayerStatus.PLAYING:
            self.pause()
        elif self.status == PlayerStatus.PAUSED:
            self.resume()

    def resume(self):
        """Resume the player"""
        if self.status == PlayerStatus.PAUSED:
            self.player.sendline("PAUSE")
            self.status = PlayerStatus.PLAYING
            self._trigger_event(MPyg321Events.USER_RESUME)
            self.on_user_resume()

    def stop(self):
        """Stops the player"""
        self.player.sendline("STOP")
        self.status = PlayerStatus.STOPPING

    def quit(self):
        """Quits the player"""
        self.player.sendline("QUIT")
        self.status = PlayerStatus.QUITTED

    def jump(self, pos):
        """Jump to position"""
        self.player.sendline("JUMP " + str(pos))

    def on_error(self):
        """Process errors encountered by the player"""
        output = self.player.readline().decode("utf-8")

        # Check error in list of errors
        for mpg_error in mpg_errors:
            if mpg_error["message"] in output:
                action = mpg_error["action"]
                context = MPyg321ErrorContext(self, action, output)
                self._trigger_event(MPyg321Events.ERROR, context)
                if action == "generic_error":
                    raise MPygError(output)
                if action == "file_error":
                    raise MPygFileError(output)
                if action == "command_error":
                    raise MPygCommandError(output)
                if action == "argument_error":
                    raise MPygArgumentError(output)
                if action == "eq_error":
                    raise MPygEQError
                if action == "seek_error":
                    raise MPygSeekError

        # Some other error occurred
        context = MPyg321ErrorContext(self, "unknown_error", output)
        self._trigger_event(MPyg321Events.ERROR, context)
        raise MPygError(output)

    def set_song(self, path):
        """song_path setter"""
        self.song_path = path

    def set_loop(self, loop):
        """loop setter"""
        self.loop = loop

    # # # Internal Callbacks # # #
    def on_music_stop_int(self):
        """Internal callback when the music is stopped"""
        if self.status == PlayerStatus.STOPPING:
            self.on_user_stop_int()
            self.status = PlayerStatus.STOPPED
        else:
            self.on_end_of_song_int()

    def on_user_stop_int(self):
        """Internal callback when the user stops the music."""
        self._trigger_event(MPyg321Events.ANY_STOP)
        self.on_any_stop()
        self._trigger_event(MPyg321Events.USER_STOP)
        self.on_user_stop()

    def on_user_pause_int(self):
        """Internal callback when user pauses the music"""
        self._trigger_event(MPyg321Events.ANY_STOP)
        self.on_any_stop()
        self._trigger_event(MPyg321Events.USER_PAUSE)
        self.on_user_pause()

    def on_user_start_or_resume_int(self):
        """Internal callback when user resumes the music"""
        self.status = PlayerStatus.PLAYING

    def on_end_of_song_int(self):
        """Internal callback when the song ends"""
        if self.loop:
            self.play()
        else:
            # The music doesn't stop if it is looped
            self._trigger_event(MPyg321Events.ANY_STOP)
            self.status = PlayerStatus.STOPPED
            self.on_any_stop()
        self._trigger_event(MPyg321Events.MUSIC_END)
        self.on_music_end()

    # # # Public Callbacks # # #
    def on_any_stop(self):
        """Callback when the music stops for any reason"""
        pass

    def on_user_pause(self):
        """Callback when user pauses the music"""
        pass

    def on_user_resume(self):
        """Callback when user resumes the music"""
        pass

    def on_user_stop(self):
        """Callback when user stops music"""
        pass

    def on_music_end(self):
        """Callback when music ends"""
        pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mpyg321",
    version="2.2.4",
    author="4br3mm0rd",
    author_email="4br3mm0rd@gmail.com",
    description="mpg321 wrapper for python - command line music player",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/4br3mm0rd/mpyg321",
    packages=setuptools.find_packages(),
    install_requires=["pexpect"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from .BasePlayer import BasePlayer
from .consts import MPyg321Events


class MPyg321Player(BasePlayer):
    """Player for legacy mpg321"""

    def __init__(
        self, player=None, audiodevice=None, performance_mode=True, custom_args=""
    ):
        self.suitable_versions = ["mpg321"]
        self.default_player = "mpg321"
        super().__init__(player, audiodevice, performance_mode, custom_args)

    def process_output_ext(self, action):
        """
        Processes specific output for mpg321 player
        It should contain the code for the mpg_out "end_of_song"
        We did not put it because the BaseClass implements a behavior
        which works for both versions
        """
        pass

    def volume(self, percent):
        """Adjust player's volume"""
        self.player.sendline("GAIN {}".format(percent))
